import time
import heapq
import logging
import fnmatch
import threading
import collections
# Py2 vs 3 compatibility.
try:
    import queue
except ImportError:
    import Queue as queue

scheduled_functions = collections.defaultdict(list)
event_handlers = collections.defaultdict(list)

# Functions that crash will have their next invocation delayed by this
# long, but only if they run more frequently than this anyway.
default_crashpenalty = 5.0

# This is a magic hacky variable that will be set by the module loader
# in fruitfly.py just before it loads a new copy of each module.
# We only need to care about this when loading multiple copies of the
# same module. See fruitfly.py's module loader for a more detailed comment.
current_modname = None

class repeat:
    """Decorator for repeatedly calling a function in a fruitfly module
    at a regular interval."""
    def __init__(self, interval):
        self._interval = float(interval)

    def __call__(self, func):
        # Add this function to the schedule.
        heapq.heappush(scheduled_functions[current_modname], (time.time() + self._interval, (func, self._interval)))

        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

class event:
    """Decorator for registering a function in a fruitfly module as a
    handler for events."""
    def __init__(self, spec):
        self._spec = spec

    def __call__(self, func):
        # Add this function to the event handlers.
        event_handlers[current_modname].append((self._spec, func))

        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

class Module(threading.Thread): 
    def __init__(self, parent_app, modname, config):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        
        self._parent_app = parent_app
        self._modname = modname
        self.config = config
        self.logger = logging.getLogger(self._modname)

        self._eventqueue = queue.Queue(maxsize = 1000)

        if getattr(self, "init", None):
            # If an init function exists, run it.
            try:
                self.init()
            except Exception as ex:
                self.logger.error("init crashed: %s (module disabled)", repr(ex))
                return

        self.start()

    def _observe_event(self, event, payload):
        self._eventqueue.put_nowait((event, payload))

    def run(self):
        while True:
            # Find out which function is scheduled next.
            try:
                (target_time, (func, interval)) = scheduled_functions[self._modname][0]
            except IndexError:
                # No scheduled functions. Set an event timeout of 100ms to prevent busy spinning.
                target_time = time.time() + 0.1
                func = None

            # Sleep until it is due to be invoked.
            timedelta = target_time - time.time()
            if timedelta > 0:
                try:
                    (event, payload) = self._eventqueue.get(timeout = timedelta)

                    # Find and execute all handlers for this event.
                    for (spec, handler) in event_handlers[self._modname]:
                        if fnmatch.fnmatch(event, spec):
                            try:
                                handler(self, event, payload)
                            except Exception as ex:
                                self.logger.error("%s handling %s event with payload '%s' crashed: %s", handler.__name__, event, payload, repr(ex))

                    continue
                except queue.Empty:
                    pass

            if func is None:
                continue

            heapq.heappop(scheduled_functions[self._modname])

            # Invoke it.
            crashpenalty = 0
            try:
                func(self)
            except Exception as ex:
                # Delay the next execution by some amount if it is scheduled
                # more frequently than the penalty amount.
                if interval < default_crashpenalty:
                    crashpenalty = default_crashpenalty

                self.logger.error("%s crashed: %s (next run delayed by %2.2fs)", func.__name__, repr(ex), crashpenalty)

            # Re-schedule it by putting it back on the heap.
            heapq.heappush(scheduled_functions[self._modname], (time.time() + interval + crashpenalty, (func, interval)))

    def send_event(self, event, payload = None):
        """Pass an event (with optional payload) up to the parent fruitfly
        instance so that it can be distributed to other modules."""
        self.logger.debug("event: %s '%s'", event, repr(payload))
        self._parent_app.send_event(event, payload)



# stdlib
import os
import imp
import glob
import time
import yaml
import logging.handlers
import inspect
from . import module

class FruitFly(object):
    def __init__(self, basedir = None, debug = False, logdestination = "stderr"):
        # Find basedir
        self._basedir = basedir or os.path.dirname(os.path.abspath(inspect.getfile(inspect.stack()[-1][0])))

        # Load config
        configpath = os.path.join(self._basedir, "fruitfly.yaml")
        self._config = yaml.safe_load(open(configpath))
        

         # Set up logging
        rootlogger = logging.getLogger()
        if self._config.get("logging", {}).get("level", "debug") == "debug":
            rootlogger.setLevel(logging.DEBUG)
        else:
            rootlogger.setLevel(logging.INFO)
            
        if self._config.get("logging", {}).get("destination", "syslog") == "debug":
            handler = logging.handlers.SysLogHandler(address = '/dev/log')
        else:
            # Writes to stderr.
            handler = logging.StreamHandler()

        handler.setFormatter(logging.Formatter('fruitfly[%(process)d] %(levelname)s %(name)s %(message)s'))
        rootlogger.addHandler(handler)
        self._logger = logging.getLogger('fruitfly')
        
        self._logger.info("Using %s to find config and modules.", self._basedir)
        self._logger.info("Loaded config from %s", configpath)
       
        # Load modules
        self._modules = {}
        self.load_modules()


    def load_modules(self):
        # Load all the modules.
        module_search_path = os.path.join(self._basedir, "mod_*.py")
        self._logger.info("Loading modules from %s", module_search_path)

        try:
            # The user provided a dict of modules.
            all_module_config = [{key: value} for key, value in self._config['modules'].items()]
        except AttributeError:
            # The user provided a list of modules.
            all_module_config = self._config['modules']

        for module_config in all_module_config:
            # Use only the first key in this dict. The key is used to specify
            # the module name.
            modname, modconfig = list(module_config.items())[0]

            # If this is not the first instance of this module, choose
            # a similar but unique name and use that instead.
            modname_internal = modname
            index = 0
            while modname_internal in self._modules:
                index += 1
                modname_internal = "%s-%d" % (modname_internal, index)

            # Set this hacky global so the 'module' module knows what
            # the name of the eventually-loaded module is going to be
            # so it can be added to the scheduler and event handlers
            # correctly. Without this, trying to load two copies of
            # one module breaks because all events go to the second
            # instance of the same module loaded.
            module.current_modname = modname_internal

            modpath = os.path.join(self._basedir, "mod_%s.py" % modname)
            self._logger.debug("Loading module %s from %s with config %s", modname, modpath, repr(modconfig))

            # Load a copy of this module.
            try:
                mod = imp.load_source(modname, modpath)
            except IOError as ex:
                self._logger.error("Config specifies module '%s', but failed to load that module. (%s)", modname, repr(ex))
                continue

            # Create an instance of the module and store it.
            self._modules[modname_internal] = getattr(mod, modname)(self, modname_internal, modconfig)

            self._logger.info("Enabled module %s", modname_internal)

    def send_event(self, event, payload):
        self._logger.debug("Distributing event %s '%s'", event, repr(payload))
        for (modname, modinst) in self._modules.items():
            modinst._observe_event(event, payload)

    def run(self):
        # Blocks forever.
        while True:
            time.sleep(1)




#!/usr/bin/env python

from distutils.core import setup

setup(
      name = 'fruitfly'
   ,  version = '0.1.4'
   ,  description = 'Modular event-driven framework.'
   ,  long_description = 'Framework for minimalist, modular, event-driven embedded applications. Intended for use on small Linux systems like the Raspberry Pi.'
   ,  author = 'Derp Ston'
   ,  author_email = 'derpston+pypi@sleepygeek.org'
   ,  url = 'https://github.com/derpston/fruitfly'
   ,  install_requires = ['PyYAML']
   ,  packages = ['fruitfly']
   )

